#!/usr/bin/env python
from __future__ import absolute_import, division, print_function

import os

import numpy as np
import pytest
import xarray as xr

from esmlab.climatology import compute_ann_mean, compute_mon_anomaly, compute_mon_climatology
from esmlab.datasets import open_dataset


@pytest.mark.parametrize("ds", ["tiny", "ccsm_pop_sample"])
def test_compute_climatology_multi(ds):
    dset = open_dataset(ds, decode_times=False)
    print(dset)
    computed_dset = compute_mon_climatology(dset)
    print(computed_dset)
    print()
    assert isinstance(computed_dset, xr.Dataset)
    computed_dset = compute_ann_mean(dset)
    print(computed_dset)
    print()
    assert isinstance(computed_dset, xr.Dataset)
    computed_dset = compute_mon_anomaly(dset)
    print(computed_dset)
    print()
    assert isinstance(computed_dset, xr.Dataset)


def test_compute_mon_climatology(dset):
    computed_dset = compute_mon_climatology(dset)
    np.testing.assert_equal(computed_dset.variable_1.values, 0.5)


def test_compute_mon_climatology_times_decoded(dset):
    computed_dset = compute_mon_climatology(dset)
    np.testing.assert_equal(computed_dset.variable_1.values, 0.5)


def test_compute_mon_anomaly(dset):
    computed_dset = compute_mon_anomaly(dset)
    assert isinstance(computed_dset, xr.Dataset)
    a = [-0.5] * 48
    b = [0.5] * 48
    a.extend(b)
    expected = np.array(a)
    np.testing.assert_equal(computed_dset.variable_1.values.ravel(), expected)


def test_compute_ann_mean(dset):
    computed_dset = compute_ann_mean(dset)
    assert isinstance(computed_dset, xr.Dataset)
    expected = np.array([0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0])
    np.testing.assert_equal(computed_dset.variable_1.values.ravel(), expected)



import sys

import numpy as np
import pytest
import xarray as xr

from esmlab.config import set_options

set_options(gridfile_directory="./tests/data")


@pytest.fixture
def dset():
    start_date = np.array([0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334])
    start_date = np.append(start_date, start_date + 365)
    end_date = np.array([31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365])
    end_date = np.append(end_date, end_date + 365)
    ds = xr.Dataset(coords={"time": 24, "lat": 2, "lon": 2, "d2": 2})
    ds["time"] = xr.DataArray(end_date, dims="time")
    ds["lat"] = xr.DataArray([0, 1], dims="lat")
    ds["lon"] = xr.DataArray([0, 1], dims="lon")
    ds["d2"] = xr.DataArray([0, 1], dims="d2")
    ds["time_bound"] = xr.DataArray(
        np.array([start_date, end_date]).transpose(), dims=["time", "d2"]
    )
    ds["variable_1"] = xr.DataArray(
        np.append(
            np.zeros([12, 2, 2], dtype="float32"), np.ones([12, 2, 2], dtype="float32"), axis=0
        ),
        dims=["time", "lat", "lon"],
    )
    ds["variable_2"] = xr.DataArray(
        np.append(
            np.ones([12, 2, 2], dtype="float32"), np.zeros([12, 2, 2], dtype="float32"), axis=0
        ),
        dims=["time", "lat", "lon"],
    )
    ds.time.attrs["units"] = "days since 0001-01-01 00:00:00"
    ds.time.attrs["calendar"] = "noleap"
    ds.time.attrs["bounds"] = "time_bound"
    return ds



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function

import numpy as np
import pytest
import xarray as xr

from esmlab import statistics

x = xr.DataArray(np.random.uniform(0, 10, (10, 10)), dims=("x", "y"), name="x_var")
y = xr.DataArray(np.random.uniform(0, 10, (10, 10)), dims=("x", "y"), name="y_var")
y[3, 3:10] = np.nan
valid = x.notnull() & y.notnull()
N = valid.sum()
x = x.where(valid)
y = y.where(valid)


def test_weighted_mean():
    with pytest.warns(UserWarning):
        w_mean = statistics.weighted_mean(x)
    np.testing.assert_allclose(w_mean, x.mean())
    assert x.attrs == w_mean.attrs
    assert x.encoding == w_mean.encoding


def test_weighted_std():
    with pytest.warns(UserWarning):
        w_std = statistics.weighted_std(x)
    np.testing.assert_allclose(w_std, x.std())
    assert x.attrs == w_std.attrs
    assert x.encoding == w_std.encoding


def test_weighted_sum():
    with pytest.warns(UserWarning):
        w_sum = statistics.weighted_sum(x)
    np.testing.assert_allclose(w_sum, x.sum())
    assert x.attrs == w_sum.attrs
    assert x.encoding == w_sum.encoding


def test_weighted_rmsd():
    rmsd = np.sqrt(((x - y) ** 2).sum() / N)
    with pytest.warns(UserWarning):
        w_rmsd = statistics.weighted_rmsd(x, y)
        np.testing.assert_allclose(rmsd, w_rmsd)


def test_weighted_cov():
    x_dev = x - x.mean()
    y_dev = y - y.mean()
    cov = (x_dev * y_dev).sum() / N
    with pytest.warns(UserWarning):
        w_cov = statistics.weighted_cov(x, y)
        np.testing.assert_allclose(cov, w_cov)


def test_weighted_corr():
    x_dev = x - x.mean()
    y_dev = y - y.mean()
    cov = (x_dev * y_dev).sum() / N
    covx = (x_dev ** 2).sum() / N
    covy = (y_dev ** 2).sum() / N
    corr = cov / np.sqrt(covx * covy)

    with pytest.warns(UserWarning):
        w_corr = statistics.weighted_corr(x, y)
        np.testing.assert_allclose(corr, w_corr)



""" Useful for:
* Testing
* building tutorials in the documentation.
"""

from __future__ import absolute_import, division, print_function

import os

import xarray as xr

try:
    from urllib.request import urlretrieve

except ImportError:
    from urllib import urlretrieve


_default_cache_dir = os.sep.join(("~", ".esmlab_data"))


def open_dataset(
    name,
    cache=True,
    cache_dir=_default_cache_dir,
    github_url="https://github.com/NCAR/esmlab-data",
    branch="master",
    **kwargs
):
    """Load a dataset from the online repository (requires access to internet).

    If a local copy is found then always use that to avoid network traffic.

    Parameters
    ----------
    name : str
        Name of the netcdf file containing the dataset
        ie. 'air_temperature'
    cache_dir : string, optional
        The directory in which to search for and write cached data.
    cache : boolean, optional
        If True, then cache data locally for use on subsequent calls
    github_url : string
        Github repository where the data is stored
    branch : string
        The git branch to download from
    kwargs : dict, optional
        Passed to xarray.open_dataset
    """
    longdir = os.path.expanduser(cache_dir)
    fullname = name + ".nc"
    localfile = os.sep.join((longdir, fullname))
    md5name = name + ".md5"
    md5file = os.sep.join((longdir, md5name))

    if not os.path.exists(localfile):

        # This will always leave this directory on disk.
        if not os.path.isdir(longdir):
            os.mkdir(longdir)

        url = "/".join((github_url, "raw", branch, fullname))
        urlretrieve(url, localfile)
        url = "/".join((github_url, "raw", branch, md5name))
        urlretrieve(url, md5file)

        localmd5 = xr.tutorial.file_md5_checksum(localfile)
        with open(md5file, "r") as f:
            remotemd5 = f.read()

        if localmd5 != remotemd5:
            os.remove(localfile)
            msg = """
            MD5 checksum does not match, try downloading dataset again.
            """
            raise IOError(msg)

    ds = xr.open_dataset(localfile, **kwargs)

    if not cache:
        ds = ds.load()
        os.remove(localfile)

    return ds



from __future__ import absolute_import, division, print_function

import xarray as xr

try:
    from contextlib import ContextDecorator
except ImportError:
    from contextlib2 import ContextDecorator


class esmlab_xr_set_options(ContextDecorator):
    """ Enables a context manager to also be used as a decorator"""

    def __init__(self, **kwargs):
        self.old = xr.set_options(**kwargs).old

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        xr.set_options(**self.old)
        return



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function

from warnings import warn

import numpy as np
import xarray as xr

from .utils.common import esmlab_xr_set_options
from .utils.variables import (
    get_original_attrs,
    get_static_variables,
    get_variables,
    save_metadata,
    set_metadata,
    set_static_variables,
    update_attrs,
)


def _apply_nan_mask(weights, x, y=None):
    # If y is specified, make sure x and y have same shape
    if y is not None and isinstance(y, xr.DataArray):
        assert x.shape == y.shape
        valid = x.notnull() & y.notnull()
    else:
        valid = x.notnull()

    # Apply nan mask
    return weights.where(valid)


def _get_weights_and_dims(x, y=None, weights=None, dim=None, apply_nan_mask=True):
    """ Get weights and dimensions """

    if dim and isinstance(dim, str):
        dims = [dim]

    elif isinstance(dim, list):
        dims = dim

    else:
        dims = [k for k in x.dims]

    op_over_dims = [k for k in dims if k in x.dims]
    if not op_over_dims:
        raise ValueError("Unexpected dimensions for variable {0}".format(x.name))

    dims_shape = tuple(l for i, l in enumerate(x.shape) if x.dims[i] in op_over_dims)
    if weights is None:
        weights = xr.DataArray(np.ones(dims_shape), dims=op_over_dims)
        weights = _apply_nan_mask(weights, x, y)

    else:
        assert weights.shape == dims_shape
        if apply_nan_mask:
            weights = _apply_nan_mask(weights, x, y)

    # Make sure weights add up to 1.0
    np.testing.assert_allclose((weights / weights.sum(op_over_dims)).sum(op_over_dims), 1.0)
    return weights, op_over_dims


@esmlab_xr_set_options(arithmetic_join="exact")
def weighted_sum(x, weights=None, dim=None, apply_nan_mask=True):
    """Reduce `xarray.DataArray` by applying `weighted sum` along some dimension(s).

            Parameters
            ----------

            x : `xarray.DataArray`
               xarray object for which to compute `weighted sum`.

            weights : array_like, optional
                    weights to use. By default, weights=`None`

            dim : str or sequence of str, optional
                Dimension(s) over which to apply mean. By default `weighted sum`
                is applied over all dimensions.

            apply_nan_mask : bool, default: True

            Returns
            -------

            Weighted_sum : `xarray.DataArray`
                New DataArray object with `weighted sum` applied to its data
                and the indicated dimension(s) removed. If `weights` is None,
                returns regular sum using equal weights for all data points.
    """
    if weights is None:
        warn("Computing sum using equal weights for all data points")

    weights, op_over_dims = _get_weights_and_dims(
        x, weights=weights, dim=dim, apply_nan_mask=apply_nan_mask
    )
    x_w_sum = (x * weights).sum(op_over_dims)

    original_attrs, original_encoding = get_original_attrs(x)
    return update_attrs(x_w_sum, original_attrs, original_encoding)


@esmlab_xr_set_options(arithmetic_join="exact")
def weighted_mean(x, weights=None, dim=None, apply_nan_mask=True):
    """Reduce `xarray.DataArray` by applying weighted mean along some dimension(s).

        Parameters
        ----------

        x : `xarray.DataArray`
           xarray object for which to compute `weighted mean`.

        weights : array_like, optional
                weights to use. By default, weights=`None`

        dim : str or sequence of str, optional
           Dimension(s) over which to apply `weighted mean`. By default weighted mean
           is applied over all dimensions.

        apply_nan_mask : bool, default: True


        Returns
        -------

        weighted_mean : `xarray.DataArray`
             New DataArray object with ` weighted mean` applied to its data
             and the indicated dimension(s) removed. If `weights` is None,
                returns regular mean using equal weights for all data points.
    """
    if weights is None:
        warn("Computing mean using equal weights for all data points")

    weights, op_over_dims = _get_weights_and_dims(
        x, weights=weights, dim=dim, apply_nan_mask=apply_nan_mask
    )

    # If the mask is applied in previous operation,
    # disable it for subseqent operations
    if apply_nan_mask:
        apply_nan_mask_flag = False
    else:
        apply_nan_mask_flag = True
    x_w_mean = weighted_sum(
        x, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag
    ) / weights.sum(op_over_dims)
    original_attrs, original_encoding = get_original_attrs(x)
    return update_attrs(x_w_mean, original_attrs, original_encoding)


@esmlab_xr_set_options(arithmetic_join="exact")
def weighted_std(x, weights=None, dim=None, ddof=0, apply_nan_mask=True):
    """Reduce `xarray.DataArray` by applying `weighted std` along some dimension(s).

        Parameters
        ----------

        x : `xarray.DataArray`
           xarray object for which to compute `weighted std`.

        weights : array_like, optional
                weights to use. By default, weights=`None`

        dim : str or sequence of str, optional
           Dimension(s) over which to apply mean. By default `weighted std`
           is applied over all dimensions.


        ddof : int, optional
            Means Delta Degrees of Freedom. By default ddof is zero.

        apply_nan_mask : bool, default: True

        Returns
        -------

        weighted_standard_deviation : `xarray.DataArray`
             New DataArray object with `weighted std` applied to its data
             and the indicated dimension(s) removed. If `weights` is None,
                returns regular standard deviation using equal weights for all data points.
    """
    if weights is None:
        warn("Computing standard deviation using equal weights for all data points")

    weights, op_over_dims = _get_weights_and_dims(
        x, weights=weights, dim=dim, apply_nan_mask=apply_nan_mask
    )

    # If the mask is applied in previous operation,
    # disable it for subseqent operations
    if apply_nan_mask:
        apply_nan_mask_flag = False
    else:
        apply_nan_mask_flag = True

    x_w_mean = weighted_mean(
        x, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag
    )

    x_w_std = np.sqrt(
        (weights * (x - x_w_mean) ** 2).sum(op_over_dims) / (weights.sum(op_over_dims) - ddof)
    )
    original_attrs, original_encoding = get_original_attrs(x)

    return update_attrs(x_w_std, original_attrs, original_encoding)


@esmlab_xr_set_options(arithmetic_join="exact")
def weighted_rmsd(x, y, weights=None, dim=None, apply_nan_mask=True):
    """ Compute weighted root-mean-square-deviation between two `xarray.DataArray` objects.

    Parameters
    ----------

    x, y : `xarray.DataArray` objects
        xarray objects for which to compute `weighted_rmsd`.

    weights : array_like, optional
                weights to use. By default, weights=`None`

    dim : str or sequence of str, optional
           Dimension(s) over which to apply `weighted rmsd` By default weighted rmsd
           is applied over all dimensions.

    apply_nan_mask : bool, default: True

    Returns
    -------

    weighted_root_mean_square deviation : float
            If `weights` is None, returns root mean square deviation using equal weights for all data points.

    """

    if weights is None:
        warn("Computing root-mean-square-deviation using equal weights for all data points")

    weights, op_over_dims = _get_weights_and_dims(
        x, weights=weights, dim=dim, apply_nan_mask=apply_nan_mask
    )

    # If the mask is applied in previous operation,
    # disable it for subseqent operations to speed up computation
    if apply_nan_mask:
        apply_nan_mask_flag = False
    else:
        apply_nan_mask_flag = True

    dev = (x - y) ** 2
    dev_mean = weighted_mean(
        dev, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag
    )
    return np.sqrt(dev_mean)


@esmlab_xr_set_options(arithmetic_join="exact")
def weighted_cov(x, y, weights=None, dim=None, apply_nan_mask=True):
    """ Compute weighted covariance between two `xarray.DataArray` objects.

    Parameters
    ----------

    x, y : `xarray.DataArray` objects
        xarray objects for which to compute `weighted covariance`.

    weights : array_like, optional
                weights to use. By default, weights=`None`

    dim : str or sequence of str, optional
           Dimension(s) over which to apply `weighted covariance`
           By default weighted covariance is applied over all dimensions.

    apply_nan_mask : bool, default: True

    Returns
    -------

    weighted_covariance : float
            If `weights` is None, returns covariance using equal weights for all data points.


    """
    if weights is None:
        warn("Computing weighted covariance using equal weights for all data points")

    weights, op_over_dims = _get_weights_and_dims(
        x, weights=weights, dim=dim, apply_nan_mask=apply_nan_mask
    )

    # If the mask is applied in previous operation,
    # disable it for subseqent operations to speed up computation
    if apply_nan_mask:
        apply_nan_mask_flag = False
    else:
        apply_nan_mask_flag = True

    mean_x = weighted_mean(x, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag)
    mean_y = weighted_mean(y, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag)

    dev_x = x - mean_x
    dev_y = y - mean_y
    dev_xy = dev_x * dev_y
    cov_xy = weighted_mean(
        dev_xy, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag
    )
    return cov_xy


@esmlab_xr_set_options(arithmetic_join="exact")
def weighted_corr(x, y, weights=None, dim=None, apply_nan_mask=True):
    """ Compute weighted correlation between two `xarray.DataArray` objects.

    Parameters
    ----------

    x, y : `xarray.DataArray` objects
        xarray objects for which to compute `weighted correlation`.

    weights : array_like, optional
             weights to use. By default, weights=`None`

    dim : str or sequence of str, optional
           Dimension(s) over which to apply `weighted correlation`
           By default weighted correlation is applied over all dimensions.

    apply_nan_mask : bool, default: True

    Returns
    -------

    weighted_correlation : float
              If `weights` is None, returns correlation using equal weights for all data points.


    """
    if weights is None:
        warn("Computing weighted correlation using equal weights for all data points")

    weights, op_over_dims = _get_weights_and_dims(
        x, weights=weights, dim=dim, apply_nan_mask=apply_nan_mask
    )

    # If the mask is applied in previous operation,
    # disable it for subseqent operations to speed up computation
    if apply_nan_mask:
        apply_nan_mask_flag = False
    else:
        apply_nan_mask_flag = True

    numerator = weighted_cov(
        x=x, y=y, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag
    )
    denominator = np.sqrt(
        weighted_cov(x, x, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag)
        * weighted_cov(y, y, weights=weights, dim=op_over_dims, apply_nan_mask=apply_nan_mask_flag)
    )
    corr_xy = numerator / denominator
    return corr_xy



import cftime
import numpy as np
import pandas as pd
import pytest
import xarray as xr

from esmlab import utils


def test_time_bound_var(dset, time_coord_name="time"):
    tm = utils.time.time_manager(dset, time_coord_name)
    results = (tm.tb_name, tm.tb_dim)
    expected = ("time_bound", "d2")
    assert results == expected


def test_time_year_to_midyeardate(dset, time_coord_name="time"):
    assert isinstance(dset[time_coord_name].values[0], np.number)
    dset = utils.time.time_year_to_midyeardate(dset, time_coord_name)
    assert isinstance(dset[time_coord_name].values[0], cftime.datetime)


def test_get_time_attrs(dset, time_coord_name="time"):
    expected = {
        "units": "days since 0001-01-01 00:00:00",
        "calendar": "noleap",
        "bounds": "time_bound",
    }
    results = utils.time.time_manager(dset, time_coord_name).time_attrs
    assert results == expected


def test_compute_time_var(dset, time_coord_name="time"):
    idx = dset.indexes[time_coord_name]
    assert isinstance(idx, pd.core.indexes.numeric.Index)
    tm = utils.time.time_manager(dset)
    results = tm.get_time_decoded()
    assert isinstance(results, xr.DataArray)


def test_uncompute_time_var(dset, time_coord_name="time"):
    dset = utils.time.compute_time_var(dset)
    assert dset[time_coord_name].dtype == np.dtype("O")
    dset_with_uncomputed_time = utils.time.uncompute_time_var(dset)
    assert np.issubdtype(dset_with_uncomputed_time[time_coord_name].dtype, np.number)


def test_sel_time(dset):
    dset = utils.time.sel_time(
        dset, indexer_val=slice("1850-01-01", "1850-12-31"), year_offset=1850
    )
    assert len(dset.time) == 12



#!/usr/bin/env python

import os
import sys

import numpy as np
import pytest
import xarray as xr

from esmlab.datasets import open_dataset

if sys.version_info[0] > 2:
    from esmlab.regrid import regridder
else:
    pass


@pytest.mark.skipif(sys.version_info[0] < 3, reason="requires python3")
def test_regrid_init():

    R = regridder(
        name_grid_src="T62", name_grid_dst="CAM_f09", method="bilinear", overwrite_existing=True
    )

    assert isinstance(R, regridder)


@pytest.mark.skipif(sys.version_info[0] < 3, reason="requires python3")
def test_regrid_regrid():

    R = regridder(
        name_grid_src="T62", name_grid_dst="CAM_f09", method="bilinear", overwrite_existing=False
    )
    ds = open_dataset(name="ncep_forecast_tseries")
    dao = R(ds.t_10)
    print(dao)
    assert isinstance(dao, xr.DataArray)



from __future__ import absolute_import, division, print_function

from datetime import datetime

import cftime
import numpy as np
import xarray as xr


def get_static_variables(dset, time_coord_name):
    return [v for v in dset.variables if time_coord_name not in dset[v].dims]


def get_variables(dset, time_coord_name, tb_name=None):
    return [
        v
        for v in dset.variables
        if time_coord_name in dset[v].dims and v not in [time_coord_name, tb_name]
    ]


def save_metadata(dset):
    attrs = {v: dset[v].attrs for v in dset.variables}
    encoding = {
        v: {
            key: val
            for key, val in dset[v].encoding.items()
            if key in ["dtype", "_FillValue", "missing_value"]
        }
        for v in dset.variables
    }
    return attrs, encoding


def set_metadata(dset, attrs, encoding, additional_attrs):
    attrs.update(additional_attrs)

    # put the attributes back
    for v in dset.variables:
        dset[v].attrs = attrs[v]

    # put the encoding back
    for v in dset.variables:
        if v in encoding:
            if dset[v].dtype == "int64":  # int64 breaks some other tools
                encoding[v]["dtype"] = "int32"
            dset[v].encoding = encoding[v]

    return dset


def set_static_variables(computed_dset, dset, static_variables):
    return xr.merge(
        (computed_dset, dset.drop([v for v in dset.variables if v not in static_variables]))
    )


def get_original_attrs(x):
    attrs = x.attrs.copy()
    encoding = x.encoding
    if "_FillValue" not in encoding:
        encoding["_FillValue"] = None
    return attrs, encoding


def update_attrs(x, original_attrs, original_encoding):
    for att in ["grid_loc", "coordinates"]:
        if att in original_attrs:
            del original_attrs[att]

    x.attrs = original_attrs
    x.encoding = {
        key: val for key, val in original_encoding.items() if key in ["_FillValue", "dtype"]
    }
    return x



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function

import xarray as xr

from esmlab.datasets import open_dataset


def test_open_dataset():
    ds = open_dataset("ccsm_pop_sample")
    assert isinstance(ds, xr.Dataset)



from __future__ import absolute_import, division, print_function

import uuid
from datetime import datetime

import cftime
import numpy as np
import xarray as xr

from .variables import get_variables

DEFAULT_TIME_UNITS = "days since 0001-01-01 00:00:00"
DEFAULT_CALENDAR = "standard"


class time_manager(object):
    """Class to support managing the time coordinate of datasets.
    """

    def __init__(self, ds, time_coord_name=None, year_offset=None):
        """
        Parameters
        ----------
        ds : xarray.Dataset
           The dataset on which to operate.
        time_coord_name : string, optional
           Name of the time coordinate of `ds`; if not provided, the
           code will attempt to infer it.
        """
        self._ds = xr.Dataset(ds)
        self.orig_time_coord_name = None
        self.orig_time_coord_decoded = None
        self._time_computed = False

        if time_coord_name is None:
            self.time_coord_name = self._infer_time_coord_name()
        else:
            self.time_coord_name = time_coord_name

        self.year_offset = year_offset

        self._infer_time_bound_var()

    @property
    def time(self):
        return self._ds[self.time_coord_name]

    @property
    def time_attrs(self):
        """Get the attributes of the time coordinate.
        """
        attrs = self.time.attrs
        encoding = self.time.encoding

        if "units" in attrs:
            units = attrs["units"]
        elif "units" in encoding:
            units = encoding["units"]
        else:
            units = None

        if "calendar" in attrs:
            calendar = attrs["calendar"]
        elif "calendar" in encoding:
            calendar = encoding["calendar"]
        else:
            calendar = DEFAULT_CALENDAR

        if "bounds" in attrs:
            bounds = attrs["bounds"]
        elif "bounds" in encoding:
            bounds = encoding["bounds"]
        else:
            bounds = None

        return {"units": units, "calendar": calendar, "bounds": bounds}

    def _infer_time_coord_name(self):
        """Infer name for time coordinate in a dataset
        """

        if "time" in self._ds.variables:
            return "time"

        unlimited_dims = self._ds.encoding.get("unlimited_dims", None)
        if len(unlimited_dims) == 1:
            return list(unlimited_dims)[0]
        raise ValueError(
            "Cannot infer `time_coord_name` from multiple unlimited dimensions: %s \n\t\t ***** Please specify time_coord_name to use. *****"
            % unlimited_dims
        )

    def _infer_time_bound_var(self):
        """Infer time_bound variable in a dataset.
        """
        self.tb_name = self.time_attrs["bounds"]
        self.tb_dim = None

        if self.tb_name:
            self.tb_dim = self._ds[self.tb_name].dims[1]

    def get_time_undecoded(self):
        """Return time undecoded.
        """
        if self.time.dtype != np.dtype("O"):
            if self.orig_time_coord_decoded is None:
                self.orig_time_coord_decoded = False
            return self.time

        if not self.time_attrs["units"]:
            raise ValueError("cannot undecode time")

        # un-decode time
        time = xr.DataArray(self.time)
        time.values = cftime.date2num(
            self.time, units=self.time_attrs["units"], calendar=self.time_attrs["calendar"]
        )
        if self.orig_time_coord_decoded is None:
            self.orig_time_coord_decoded = True
        return time

    def get_time_decoded(self, midpoint=True):
        """Return time decoded.
        """
        # to compute a time midpoint, we need a time_bound variable
        if midpoint and self.time_bound is None:
            raise ValueError("cannot compute time midpoint w/o time bounds")

        if midpoint:
            time_values = self.time_bound.mean(self.tb_dim)

        else:
            # if time has already been decoded and there's no year_offset,
            # just return the time as is
            if self.time.dtype == np.dtype("O"):
                if self.year_offset is None:
                    return time_values

                # if we need to un-decode time to apply the year_offset,
                # make sure there are units to do so with
                time_values = self.get_time_undecoded()

            # time has not been decoded
            else:
                time_values = self.time

        if self.year_offset is not None:
            time_values += cftime.date2num(
                datetime(int(self.year_offset), 1, 1),
                units=self.time_attrs["units"],
                calendar=self.time_attrs["calendar"],
            )

        time_out = xr.DataArray(self.time)
        time_out.values = xr.CFTimeIndex(
            cftime.num2date(
                time_values, units=self.time_attrs["units"], calendar=self.time_attrs["calendar"]
            )
        )
        return time_out

    def compute_time(self):
        """Compute the mid-point of the time bounds.
        """

        if self._time_computed:
            return self._ds

        self.orig_time_coord_name = "t" + uuid.uuid4().hex

        self._ds[self.orig_time_coord_name] = self.get_time_undecoded()

        if self.time_bound is not None:
            groupby_coord = self.get_time_decoded(midpoint=True)

        else:
            groupby_coord = self.get_time_decoded(midpoint=False)

        self._ds[self.time_coord_name].values = groupby_coord.values

        self._time_computed = True

        return self._ds

    def restore_dataset(self, ds):
        """Return the original time variable.
        """
        if not self._time_computed:
            raise ValueError("time was not computed; cannot restore dataset")
        time_values = ds[self.orig_time_coord_name].values
        if self.orig_time_coord_decoded:
            time_values = xr.CFTimeIndex(
                cftime.num2date(
                    time_values,
                    units=self.time_attrs["units"],
                    calendar=self.time_attrs["calendar"],
                )
            )
        ds[self.time_coord_name].values = time_values
        ds = ds.drop(self.orig_time_coord_name)
        return ds

    @property
    def time_bound(self):
        """return time bound, ensuring that it has not been decoded.
        """
        if self.tb_name is None:
            return

        if self._ds[self.tb_name].dtype == np.dtype("O"):
            tb_value = cftime.date2num(
                self._ds[self.tb_name],
                units=self.time_attrs["units"],
                calendar=self.time_attrs["calendar"],
            )
            return xr.DataArray(tb_value, dims=(self.time_coord_name))
        else:
            return self._ds[self.tb_name]

    @property
    def time_bound_diff(self):
        """Compute the difference between time bounds.
        """
        time_bound_diff = xr.ones_like(self.time, dtype=self.time.dtype)

        time_bound_diff.name = self.tb_name + "_diff"
        time_bound_diff.attrs = {}

        if self.time_bound is not None:
            time_bound_diff.values = self.time_bound.diff(dim=self.tb_dim)[:, 0]
            if self.tb_dim in time_bound_diff.coords:
                time_bound_diff = time_bound_diff.drop(self.tb_dim)

        return time_bound_diff


def time_year_to_midyeardate(ds, time_coord_name):
    """Set the time coordinate to the mid-point of the year.
    """
    ds[time_coord_name].values = np.array(
        [cftime.datetime(year, 7, 2) for year in ds[time_coord_name]]
    )
    return ds


def compute_time_var(ds, midpoint=True, year_offset=None):
    """Compute the time coordinate of a dataset.

    Parameters
    ----------
    ds : `xarray.Dataset`
       The dataset on which to operate
    midpoint : bool, optional [default=True]
       Return time at the midpoints of the `time:bounds`
    year_offset : numeric, optional
       Integer year by which to offset the time axis.

    Returns
    -------
    ds : `xarray.Dataset`
       The dataset with time coordinate modified.
    """
    tm = time_manager(ds, year_offset=year_offset)
    ds = tm.compute_time()
    ds[tm.time_coord_name] = tm.get_time_decoded(midpoint)
    ds = ds.drop(tm.orig_time_coord_name)
    return ds


def uncompute_time_var(ds):
    """Return time coordinate from object to float.

    Parameters
    ----------
    ds : `xarray.Dataset`
       The dataset on which to operate

    Returns
    -------
    ds : `xarray.Dataset`
       The dataset with time coordinate modified.
    """
    tm = time_manager(ds)
    ds = tm.compute_time()
    ds[tm.time_coord_name] = tm.get_time_undecoded()
    ds = ds.drop(tm.orig_time_coord_name)
    return ds


def sel_time(ds, indexer_val, year_offset=None):
    """Return dataset truncated to specified time range.

    Parameters
    ----------
    ds : `xarray.Dataset`
       The dataset on which to operate
    indexer_val : scalar, slice, or array_like
       value passed to ds.isel(**{time_coord_name: indexer_val})
    year_offset : numeric, optional
       Integer year by which to offset the time axis.

    Returns
    -------
    ds : `xarray.Dataset`
       The dataset with time coordinate truncated.
    """

    tm = time_manager(ds, year_offset=year_offset)
    ds = tm.compute_time()
    ds = ds.sel(**{tm.time_coord_name: indexer_val})
    return tm.restore_dataset(ds)



""" The configuration script: set global settings.
"""

from __future__ import absolute_import, print_function

import os

import yaml

GRIDFILE_DIRECTORY = "gridfile_directory"
GRID_DEFITIONS_FILE = "grid_defitions_file"

_here = os.path.abspath(os.path.dirname(__file__))
_config_dir = os.path.join(os.path.expanduser("~"), ".esmlab")
_path_config_yml = os.path.join(_config_dir, "config.yml")

if os.path.exists(".config-esmlab.yml"):
    _path_config_yml = os.path.join(".config-esmlab.yml")

SETTINGS = {
    GRIDFILE_DIRECTORY: os.path.join(_config_dir, "esmlab-grid-files"),
    GRID_DEFITIONS_FILE: os.path.join(_here, "grid_definitions.yml"),
}

for key in [GRIDFILE_DIRECTORY]:
    if not os.path.exists(SETTINGS[key]):
        os.makedirs(SETTINGS[key])


def _check_path_write_access(value):
    value = os.path.abspath(os.path.expanduser(value))
    if os.path.exists(value):
        if not os.access(value, os.W_OK):
            print("no write access to: {0}".format(value))
            return False
        return True

    try:
        os.makedirs(value)
        return True
    except (OSError, PermissionError) as err:
        print("could not make directory: {0}".format(value))
        raise err


def _full_path(value):
    return os.path.abspath(os.path.expanduser(value))


def _check_exists(value):
    return os.path.exists(value)


_VALIDATORS = {GRIDFILE_DIRECTORY: _check_path_write_access, GRID_DEFITIONS_FILE: _check_exists}

_SETTERS = {GRIDFILE_DIRECTORY: _full_path}


class set_options(object):
    """Set configurable settings."""

    def __init__(self, **kwargs):
        self.old = {}
        for key, val in kwargs.items():
            if key not in SETTINGS:
                raise ValueError(
                    "{key} is not in the set of valid settings:\n {set}".format(
                        key=key, set=set(SETTINGS)
                    )
                )
            if key in _VALIDATORS and not _VALIDATORS[key](val):
                raise ValueError("{val} is not a valid value for {key}".format(key=key, val=val))
            self.old[key] = SETTINGS[key]
        self._apply_update(kwargs)

    def _apply_update(self, settings_dict):
        for key, val in settings_dict.items():
            if key in _SETTERS:
                settings_dict[key] = _SETTERS[key](val)
        SETTINGS.update(settings_dict)

    def __enter__(self):
        return

    def __exit__(self, type, value, traceback):
        self._apply_update(self.old)


def get_options():
    return SETTINGS


if os.path.exists(_path_config_yml):
    with open(_path_config_yml) as f:
        dot_file_settings = yaml.load(f)
    if dot_file_settings:
        set_options(**dot_file_settings)



#!/usr/bin/env python
"""Top-level module for esmlab."""


import sys

from . import climatology, datasets, statistics
from ._version import get_versions
from .config import get_options, set_options

if sys.version_info > (3, 0):
    from .regrid import regridder

__version__ = get_versions()["version"]
del get_versions


_module_imports = (climatology, statistics, datasets, set_options, get_options)
__all__ = list(map(lambda x: x.__name__, _module_imports))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import sys
from setuptools import setup, find_packages
import versioneer
from os.path import exists

if exists('README.rst'):
    with open('README.rst') as f:
        long_description = f.read()
else:
    long_description = ''

if sys.version_info > (3, 0):
    with open('requirements.txt') as f:
        install_requires = f.read().strip().split('\n')
else:
    with open('requirements-py2.txt') as f:
        install_requires = f.read().strip().split('\n')

test_requirements = ['pytest']

setup(
    maintainer='Anderson Banihirwe',
    maintainer_email='abanihi@ucar.edu',
    description='Tools for working with earth system multi-model analyses with xarray',
    install_requires=install_requires,
    license='Apache License 2.0',
    long_description=long_description,
    keywords='esmlab xarray cmip cesm',
    name='esmlab',
    packages=find_packages(),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/NCAR/esmlab',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
)



import logging
import os

import ESMF
import numpy as np
import xarray as xr
import xesmf as xe

from .config import SETTINGS

logger = logging.getLogger(__name__)


class _grid_ref(object):
    """Get ESMF grid object from grid file.
    """

    def __init__(self, name, overwrite_existing=False):
        self.name = name
        grid_file_dir = SETTINGS["gridfile_directory"]
        self.scrip_grid_file = f"{grid_file_dir}/{self.name}.nc"

        self._gen_grid_file(overwrite_existing=overwrite_existing)

        self._esmf_grid_from_scrip()

    def _gen_grid_file(self, overwrite_existing):
        """Generate a grid file.
        """
        if os.path.exists(self.scrip_grid_file) and not overwrite_existing:
            return

        raise NotImplementedError("grid file generation is not implemented")

    def _esmf_grid_from_scrip(self):
        """Generate an ESMF grid object from a SCRIP grid file.
        """

        if not os.path.exists(self.scrip_grid_file):
            raise ValueError(f"file not found: {self.scrip_grid_file}")

        self.ds = xr.open_dataset(self.scrip_grid_file)

        self.grid = ESMF.Grid(
            filename=self.scrip_grid_file,
            filetype=ESMF.api.constants.FileFormat.SCRIP,
            add_corner_stagger=True,
        )

        self.shape = self.ds.grid_dims.values
        mask = self.grid.add_item(ESMF.GridItem.MASK)
        mask[:] = self.ds.grid_imask.values.reshape(self.shape[::-1]).T


class regridder(object):
    """Class to enable regridding between named grids.
    """

    def __init__(self, name_grid_src, name_grid_dst, method="bilinear", overwrite_existing=False):
        """
        Parameters
        ----------
        name_grid_src : string
                      Name of source grid.
        name_grid_dst : string
                      Name of destination grid.
        method : string, optional
                Regridding method. Options are:

                - 'bilinear'
                - 'conservative'
                - 'patch'
                - 'nearest_s2d'
                - 'nearest_d2s'
        overwrite_existing : bool, optional [Default=False]
                           Overwrite previously generated weight files.

        """

        self.name_grid_src = name_grid_src
        self.name_grid_dst = name_grid_dst
        self.grid_ref_src = _grid_ref(name_grid_src)
        self.grid_ref_dst = _grid_ref(name_grid_dst)
        self.method = method

        self.N_src = self.grid_ref_src.shape[0] * self.grid_ref_src.shape[1]
        self.N_dst = self.grid_ref_dst.shape[0] * self.grid_ref_dst.shape[1]

        self._gen_weights(overwrite_existing=overwrite_existing)
        self.A = xe.smm.read_weights(self.weight_file, self.N_src, self.N_dst)

    def _gen_weights(self, overwrite_existing):
        """Generate regridding weights
        """
        grid_file_dir = SETTINGS["gridfile_directory"]

        self.weight_file = "{0}/{1}_to_{2}_{3}.nc".format(
            grid_file_dir, self.name_grid_src, self.name_grid_dst, self.method
        )

        if os.path.exists(self.weight_file):
            if overwrite_existing:
                logger.info(f"removing {self.weight_file}")
                os.remove(self.weight_file)
            else:
                return

        logger.info(f"generating {self.weight_file}")
        regrid = xe.backend.esmf_regrid_build(
            self.grid_ref_src.grid,
            self.grid_ref_dst.grid,
            filename=self.weight_file,
            method=self.method,
        )

        xe.backend.esmf_regrid_finalize(regrid)

    def __call__(
        self,
        da_in,
        renormalize=True,
        apply_mask=None,
        interp_coord=None,
        post_method=None,
        post_method_kwargs=None,
    ):
        """Perform regridding on an `xarray.DataArray`.

        Parameters
        ----------
        da_in : xr.DataArray
              The data to regrid
        renormalize : bool, optional [default=True]
                   Logical flag to trigger renormalization of the remapping
                   weights. This is useful if the remapping weight-file was
                   computed with a different missing value mask than `da_in`.
                   For instance, in regridding 3D ocean data, it is possible
                   to use a mapping file computed at the surface at each
                   successive depth level: setting `renormalize=True` will
                   ensure correct handling of missing values.
        interp_coord : dict, optional
                   Dictionary specifying dimension names an new coordinates;
                   passed to `xarray.DataArray.interp`.
                   New coordinate can be a scalar, array-like or DataArray.
                   If DataArrays are passed as new coordates, their dimensions
                   are used for the broadcasting.
        apply_mask : bool, optional [default=False]
                  Apply a missing-values mask after regridding operations.
        post_method : callable, optional
                 If provided, call this function on DataArray after regridding.
        post_method_kwargs : dict, optional
                  Keyword arguments to pass to `post_method`.

        Returns
        -------
        da_out : `xarray.DataArray`
            The dataset regridded to the destination grid.
        """

        return self.regrid_dataarray(
            da_in,
            renormalize=renormalize,
            apply_mask=apply_mask,
            interp_coord=interp_coord,
            post_method=post_method,
            post_method_kwargs=post_method_kwargs,
        )

    def regrid_dataarray(
        self,
        da_in,
        renormalize=True,
        interp_coord=None,
        apply_mask=None,
        post_method=None,
        post_method_kwargs=None,
    ):

        # pull data, dims and coords from incoming DataArray
        data_src = da_in.values
        non_lateral_dims = da_in.dims[:-2]
        copy_coords = {d: da_in.coords[d] for d in non_lateral_dims if d in da_in.coords}

        # if renormalize == True, remap a field of ones
        if renormalize:
            ones_src = np.where(np.isnan(data_src), 0.0, 1.0)
            data_src = np.where(np.isnan(data_src), 0.0, data_src)

        # remap the field
        data_dst = xe.smm.apply_weights(
            self.A, data_src, self.grid_ref_dst.shape[1], self.grid_ref_dst.shape[0]
        )

        # renormalize to include only non-missing data_src
        if renormalize:
            ones_dst = xe.smm.apply_weights(
                self.A, ones_src, self.grid_ref_dst.shape[1], self.grid_ref_dst.shape[0]
            )
            ones_dst = np.where(ones_dst > 0.0, ones_dst, np.nan)
            data_dst = data_dst / ones_dst
            data_dst = np.where(ones_dst > 0.0, data_dst, np.nan)

        # reform into xarray.DataArray
        da_out = xr.DataArray(
            data_dst, name=da_in.name, dims=da_in.dims, attrs=da_in.attrs, coords=copy_coords
        )
        da_out.attrs["regrid_method"] = self.method

        # interpolate coordinates (i.e., vertical)
        # setup to copy lowest/highest values where extrapolation is needed
        if interp_coord:
            for dim, new_coord in interp_coord.items():
                if dim in da_out.dims:
                    extrap_values = (da_out.isel(**{dim: 0}), da_out.isel(**{dim: -1}))

                    da_out = da_out.interp(
                        coords={dim: new_coord},
                        method="linear",
                        assume_sorted=True,
                        kwargs={"fill_value": extrap_values},
                    )

        # apply a missing-values mask
        if apply_mask is not None:
            if apply_mask.dims != da_in.dims:
                logger.warning(f"masking {apply_mask.dims}; " f"data have dims: {da_in.dims}")
            da_out = da_out.where(apply_mask)

        # apply a post_method
        if post_method is not None:
            da_out = post_method(da_out, **post_method_kwargs)

        return da_out



#!/usr/bin/env python
"""Contains functions to compute climatologies."""
from __future__ import absolute_import, division, print_function

import numpy as np
import xarray as xr

from .utils.common import esmlab_xr_set_options
from .utils.time import time_manager, time_year_to_midyeardate
from .utils.variables import (
    get_original_attrs,
    get_static_variables,
    get_variables,
    save_metadata,
    set_metadata,
    set_static_variables,
    update_attrs,
)


@esmlab_xr_set_options(arithmetic_join="exact")
def compute_mon_climatology(dset, time_coord_name=None):
    """Calculates monthly climatology (monthly means)

    Parameters
    ----------
    dset : xarray.Dataset
           The data on which to operate

    time_coord_name : string
            Name for time coordinate

    Returns
    -------
    computed_dset : xarray.Dataset
                    The computed monthly climatology data

    """

    tm = time_manager(dset, time_coord_name)
    dset = tm.compute_time()
    time_coord_name = tm.time_coord_name

    static_variables = get_static_variables(dset, time_coord_name)

    # save metadata
    attrs, encoding = save_metadata(dset)

    # Compute climatology
    time_dot_month = ".".join([time_coord_name, "month"])
    computed_dset = (
        dset.drop(static_variables)
        .groupby(time_dot_month)
        .mean(time_coord_name)
        .rename({"month": time_coord_name})
    )

    # Put static_variables back
    computed_dset = set_static_variables(computed_dset, dset, static_variables)

    # add month_bounds
    computed_dset["month"] = computed_dset[time_coord_name].copy()
    attrs["month"] = {"long_name": "Month", "units": "month"}
    encoding["month"] = {"dtype": "int32", "_FillValue": None}
    encoding[time_coord_name] = {"dtype": "float", "_FillValue": None}

    if "calendar" in attrs[time_coord_name]:
        attrs[time_coord_name]["calendar"] = attrs[time_coord_name]["calendar"]

    if tm.time_bound is not None:
        computed_dset[tm.tb_name] = tm.time_bound - tm.time_bound[0, 0]
        computed_dset[time_coord_name].values = computed_dset[tm.tb_name].mean(tm.tb_dim).values

        encoding[tm.tb_name] = {"dtype": "float", "_FillValue": None}
        attrs[tm.tb_name] = {"long_name": tm.tb_name, "units": "days since 0001-01-01 00:00:00"}

        attrs[time_coord_name] = {
            "long_name": time_coord_name,
            "units": "days since 0001-01-01 00:00:00",
            "bounds": tm.tb_name,
        }

        if "calendar" in attrs[time_coord_name]:
            attrs[tm.tb_name]["calendar"] = attrs[time_coord_name]["calendar"]

    # Put the attributes, encoding back
    computed_dset = set_metadata(computed_dset, attrs, encoding, additional_attrs={})

    computed_dset = tm.restore_dataset(computed_dset)

    return computed_dset


@esmlab_xr_set_options(arithmetic_join="exact")
def compute_mon_anomaly(dset, slice_mon_clim_time=None, time_coord_name=None):
    """Calculates monthly anomaly

    Parameters
    ----------
    dset : xarray.Dataset
           The data on which to operate

    slice_mon_clim_time : slice, optional
                          a slice object passed to
                          `dset.isel(time=slice_mon_clim_time)` for subseting
                          the time-period overwhich the climatology is computed
    time_coord_name : string
            Name for time coordinate

    Returns
    -------
    computed_dset : xarray.Dataset
                    The computed monthly anomaly data

    """

    tm = time_manager(dset, time_coord_name)
    dset = tm.compute_time()
    time_coord_name = tm.time_coord_name

    static_variables = get_static_variables(dset, time_coord_name)

    # save metadata
    attrs, encoding = save_metadata(dset)

    # Compute anomaly
    time_dot_month = ".".join([time_coord_name, "month"])
    if slice_mon_clim_time is None:
        computed_dset = dset.drop(static_variables).groupby(time_dot_month) - dset.drop(
            static_variables
        ).groupby(time_dot_month).mean(time_coord_name)
    else:
        computed_dset = dset.drop(static_variables).groupby(time_dot_month) - dset.drop(
            static_variables
        ).sel(time=slice_mon_clim_time).groupby(time_dot_month).mean(time_coord_name)

    # reset month to become a variable
    computed_dset = computed_dset.reset_coords("month")

    # Put static_variables back
    computed_dset = set_static_variables(computed_dset, dset, static_variables)

    # Put the attributes, encoding back
    computed_dset = set_metadata(
        computed_dset, attrs, encoding, additional_attrs={"month": {"long_name": "Month"}}
    )

    computed_dset = tm.restore_dataset(computed_dset)

    return computed_dset


@esmlab_xr_set_options(arithmetic_join="exact")
def compute_ann_mean(dset, weights=None, time_coord_name=None):
    """Calculates annual climatology (annual means)

    Parameters
    ----------
    dset : xarray.Dataset
           The data on which to operate

    weights : array_like, optional
              weights to use for each time period.
              If None and dataset doesn't have `time_bound` variable,
              every time period has equal weight of 1.

    time_coord_name : string
            Name for time coordinate

    Returns
    -------
    computed_dset : xarray.Dataset
                    The computed annual climatology data

    """

    tm = time_manager(dset, time_coord_name)
    dset = tm.compute_time()
    time_coord_name = tm.time_coord_name

    static_variables = get_static_variables(dset, time_coord_name)
    variables = get_variables(dset, time_coord_name, tm.tb_name)
    # save metadata
    attrs, encoding = save_metadata(dset)

    time_dot_year = ".".join([time_coord_name, "year"])
    # Compute weights
    if weights:
        if len(weights) != len(dset[time_coord_name]):
            raise ValueError("weights and dataset time values must be of the same length")
        else:
            dt = xr.ones_like(dset[time_coord_name], dtype=bool)
            dt.values = weights
            weights = dt / dt.sum(xr.ALL_DIMS)
            np.testing.assert_allclose(weights.sum(xr.ALL_DIMS), 1.0)

    elif not weights:
        dt = tm.time_bound_diff

        weights = dt.groupby(time_dot_year) / dt.groupby(time_dot_year).sum(xr.ALL_DIMS)

        np.testing.assert_allclose(weights.groupby(time_dot_year).sum(xr.ALL_DIMS), 1.0)

    # groupby.sum() does not seem to handle missing values correctly: yields 0 not nan
    # the groupby.mean() does return nans, so create a mask of valid values
    # for each variable
    valid = {
        v: dset[v]
        .groupby(time_dot_year)
        .mean(dim=time_coord_name)
        .notnull()
        .rename({"year": time_coord_name})
        for v in variables
    }
    ones = (
        dset.drop(static_variables)
        .where(dset.drop(static_variables).isnull())
        .fillna(1.0)
        .where(dset.drop(static_variables).notnull())
        .fillna(0.0)
    )

    # Compute annual mean
    computed_dset = (
        (dset.drop(static_variables) * weights)
        .groupby(time_dot_year)
        .sum(time_coord_name)
        .rename({"year": time_coord_name})
    )
    ones_out = (
        (ones * weights)
        .groupby(time_dot_year)
        .sum(time_coord_name)
        .rename({"year": time_coord_name})
    )
    ones_out = ones_out.where(ones_out > 0.0)

    # Renormalize to appropriately account for missing values
    computed_dset = computed_dset / ones_out

    # Apply the valid-values mask
    for v in variables:
        computed_dset[v] = computed_dset[v].where(valid[v])

    # address time
    attrs[time_coord_name] = {
        "long_name": time_coord_name,
        "units": "days since 0001-01-01 00:00:00",
    }

    if "calendar" in attrs[time_coord_name]:
        attrs[time_coord_name]["calendar"] = attrs[time_coord_name]["calendar"]

    # compute the time_bound variable
    if tm.time_bound is not None:
        tb_out_lo = (
            tm.time_bound[:, 0]
            .groupby(time_dot_year)
            .min(dim=time_coord_name)
            .rename({"year": time_coord_name})
        )
        tb_out_hi = (
            tm.time_bound[:, 1]
            .groupby(time_dot_year)
            .max(dim=time_coord_name)
            .rename({"year": time_coord_name})
        )

        computed_dset[tm.tb_name] = xr.concat((tb_out_lo, tb_out_hi), dim=tm.tb_dim)
        attrs[time_coord_name]["bounds"] = tm.tb_name

    # Put static_variables back
    computed_dset = set_static_variables(computed_dset, dset, static_variables)

    # make year into date
    computed_dset = time_year_to_midyeardate(computed_dset, time_coord_name)

    # Put the attributes, encoding back
    computed_dset = set_metadata(computed_dset, attrs, encoding, additional_attrs={})

    computed_dset = tm.restore_dataset(computed_dset)

    return computed_dset
